from .agent import (
    Agent,
    Tool,
    ToolResult,
    create_tool,
    logger,
)
from .historymanager import HistoryManager, StaticHistoryManager

__all__ = [
    "Agent",
    "HistoryManager",
    "StaticHistoryManager",
    "Tool",
    "ToolResult",
    "create_tool",
    "logger",
]



import io
import json
import logging
import inspect
import base64
import mimetypes
import requests
import asyncio
import whisper
from termcolor import cprint
from .utils.audio import load_audio
from dataclasses import dataclass
from typing import (
    AsyncGenerator,
    Callable,
    Dict,
    List,
    Optional,
    Any,
    Union,
)
from openai import AsyncOpenAI

logger = logging.getLogger(__name__)


@dataclass
class ToolResult:
    content: Any
    output_schema: Optional[Dict[str, Any]] = None


@dataclass
class Tool:
    name: str
    description: str
    execute: Callable[..., Any]
    announcement_phrase: Union[str, Callable[[Dict[str, Any]], str]]
    input_schema: Dict[str, Any]
    output_schema: Dict[str, Any]


@dataclass
class ToolCallEvent:
    tool_name: str
    arguments: Optional[Dict[str, Any]] = None
    result: Optional[Any] = None
    announcement_phrase: Optional[str] = None


@dataclass
class ChatStreamEvent:
    content: Optional[str] = None
    reasoning: Optional[str] = None
    tool_call: Optional[ToolCallEvent] = None


def create_tool(
    id: str,
    description: str,
    execute: Callable[..., Any],
    announcement_phrase: Union[str, Callable[[Dict[str, Any]], str]],
    input_schema: Dict[str, Any],
    output_schema: Dict[str, Any],
) -> Tool:
    """Create a Tool instance from the provided metadata and callbacks."""
    return Tool(
        name=id,
        description=description,
        execute=execute,
        input_schema=input_schema,
        output_schema=output_schema,
        announcement_phrase=announcement_phrase,
    )


class Agent:
    model: str
    system_prompt: str
    max_tool_rounds: int
    stt: Optional[Any]
    tts: Optional[Any]
    base_url: str
    history: List[Dict[str, Any]]
    tools: List[Dict[str, Any]]
    available_functions: Dict[str, Callable[..., Any]]
    tool_objects: Dict[str, Tool]
    generation_params: Dict[str, Any]
    parallel_tool_calls: bool
    client: AsyncOpenAI

    def __init__(
        self,
        model: str,
        system_prompt: str,
        tools: Optional[List[Tool]] = None,
        base_url: str = "http://127.0.0.1:8080/v1",
        api_key: str = "",
        generation_params: Optional[Dict[str, Any]] = None,
        max_tool_rounds: int = 50,
        parallel_tool_calls: bool = True,
    ) -> None:
        """
        Initialize the Agent.

        Args:
            model: The model to use.
            system_prompt: The system prompt to use.
            tools: The tools to use.
            base_url: The base URL to use.
            api_key: The API key to use.
            generation_params: The generation parameters to use.
            max_tool_rounds: The maximum number of tool rounds allowed before giving up.
            parallel_tool_calls: Whether to allow parallel tool calls.
        """
        self.model = model
        self.system_prompt = system_prompt
        self.max_tool_rounds = max_tool_rounds
        self.stt = None
        self.tts = None
        self.base_url = base_url
        self.history: List[Dict[str, Any]] = []
        self.tools: List[Dict[str, Any]] = []
        self.available_functions: Dict[str, Callable] = {}
        self.tool_objects: Dict[str, Tool] = {}
        self.generation_params = generation_params or {}
        self.parallel_tool_calls = parallel_tool_calls
        if tools:
            for tool in tools:
                if not isinstance(tool, Tool):
                    self.tools.append(tool)
                    continue

                params = tool.input_schema
                self.tools.append(
                    {
                        "type": "function",
                        "function": {
                            "name": tool.name,
                            "description": tool.description,
                            "parameters": params,
                        },
                    }
                )
                self.available_functions[tool.name] = tool.execute
                self.tool_objects[tool.name] = tool

        if self.parallel_tool_calls:
            self._register_parallel_tool()

        if self.system_prompt:
            self.history.append({"role": "system", "content": self.system_prompt})

        self.client = AsyncOpenAI(base_url=base_url, api_key=api_key)
        self.show_yourself()

    def add_image_to_history(
        self, history: List[Dict[str, Any]], image_path: str
    ) -> List[Dict[str, Any]]:
        """Append a base64-encoded image message to the provided history."""
        with open(image_path, "rb") as image_file:
            base64_image = base64.b64encode(image_file.read()).decode("utf-8")
        # Determine mime type
        mime_type, _ = mimetypes.guess_type(image_path)
        if not mime_type:
            mime_type = "image/jpeg"
        history.append(
            {
                "type": "image_url",
                "image_url": {"url": f"data:{mime_type};base64,{base64_image}"},
            }
        )

        return history

    def add_voice_message_to_history(
        self, history: List[Dict[str, Any]], base64_audio_bytes: str
    ) -> List[Dict[str, Any]]:
        """Transcribe base64 audio and append it to the history as user content."""
        if not self.stt:
            self.stt = whisper.load_model("base.en")

        audio, _ = load_audio(
            io.BytesIO(base64.b64decode(base64_audio_bytes)), sr=16000, mono=True
        )
        result = self.stt.transcribe(audio)
        print(result["text"])
        history.append({"role": "user", "content": result["text"]})
        return history

    def speak(
        self,
        text: str,
        ref_text: str,
        ref_audio_path: Optional[str] = None,
        backbone_path: str = "neuphonic/neutts-nano-q4-gguf",
        codec_path: str = "neuphonic/neucodec-onnx-decoder",
    ) -> Any:
        """
        Generate TTS audio using the default NeuTTS backend.

        Args:
            text: Text to synthesize.
            ref_text: Reference text matching the reference audio.
            ref_audio_path: Path to a reference audio file to encode for voice conditioning.
        """

        if not ref_audio_path:
            raise ValueError("Provide ref_audio_path for TTS.")

        if not ref_text:
            raise ValueError("Provide ref_text for TTS.")

        if self.tts is None:
            try:
                from .neutts_minimal import NeuTTSMinimal
            except ModuleNotFoundError as exc:
                raise ModuleNotFoundError(
                    "TTS dependencies are not installed. "
                    "Install fury-sdk[tts] and ensure espeak is available."
                ) from exc
            self.tts = NeuTTSMinimal(
                backbone_path=backbone_path,
                codec_path=codec_path,
            )

        return self.tts.infer_stream(
            text=text,
            ref_audio_path=ref_audio_path,
            ref_text=ref_text,
        )

    async def chat(
        self,
        history: List[Dict[str, Any]],
        reasoning: bool = False,
    ) -> AsyncGenerator[ChatStreamEvent, None]:
        """
        Chat with the agent. Streams the response from the agent.

        Args:
            history: List of messages in the conversation.

        Returns:
            AsyncGenerator of ChatStreamEvent entries with content, reasoning, or tool data.
        """

        if not self._validate_history(history):
            raise ValueError("History is not in the correct format")

        def format_multimodal_content(res):
            if isinstance(res, dict) and "image_base64" in res:
                description = res.get("description", "Image captured from webcam.")
                user_message = {
                    "role": "user",
                    "content": [
                        {
                            "type": "text",
                            "text": description,
                        },
                        {
                            "type": "image_url",
                            "image_url": {
                                "url": f"data:image/jpeg;base64,{res['image_base64']}"
                            },
                        },
                    ],
                }
                return description, user_message
            return str(res), None

        try:
            if not reasoning and "qwen" in self.model.lower():
                history[-1]["content"] = f"{history[-1]['content']} /no_think"

            if history is not None:
                # Ensure the system prompt is included when caller provides history.
                active_history = list(history)
                if self.system_prompt and not any(
                    msg.get("role") == "system" for msg in active_history
                ):
                    active_history = [
                        {"role": "system", "content": self.system_prompt},
                        *active_history,
                    ]

            def tool_error_payload(call_id: str, name: str, msg: str):
                return (
                    {
                        "tool_call_id": call_id,
                        "role": "tool",
                        "name": name,
                        "content": msg,
                    },
                    msg,
                )

            for _ in range(self.max_tool_rounds):
                tool_calls: List[Dict[str, Any]] = []
                kwargs: Dict[str, Any] = {
                    "model": self.model,
                    "messages": active_history,
                    "stream": True,
                }
                if not reasoning:
                    kwargs["extra_body"] = {
                        "chat_template_kwargs": {"enable_thinking": False}
                    }
                if self.tools:
                    kwargs["tools"] = self.tools

                if self.generation_params:
                    kwargs.update(self.generation_params)

                completion = await self.client.chat.completions.create(**kwargs)

                async for chunk in completion:  # type: ignore
                    delta = chunk.choices[0].delta
                    if delta.tool_calls:
                        for tc_chunk in delta.tool_calls:
                            if len(tool_calls) <= tc_chunk.index:
                                tool_calls.append(
                                    {
                                        "id": "",
                                        "type": "function",
                                        "function": {"name": "", "arguments": ""},
                                    }
                                )
                            tc = tool_calls[tc_chunk.index]
                            if tc_chunk.id:
                                tc["id"] += tc_chunk.id
                            if tc_chunk.function.name:
                                tc["function"]["name"] += tc_chunk.function.name
                            if tc_chunk.function.arguments:
                                tc["function"][
                                    "arguments"
                                ] += tc_chunk.function.arguments

                    reasoning_content = getattr(delta, "reasoning_content", None)
                    if reasoning_content:
                        yield ChatStreamEvent(reasoning=reasoning_content)
                        continue

                    if delta.content:
                        yield ChatStreamEvent(content=delta.content)

                if not tool_calls:
                    return

                active_history.append({"role": "assistant", "tool_calls": tool_calls})

                for tool_call in tool_calls:
                    fname = tool_call["function"]["name"]
                    call_id = tool_call["id"]

                    if fname not in self.available_functions:
                        continue

                    try:
                        args = json.loads(tool_call["function"]["arguments"])
                    except json.JSONDecodeError as e:
                        payload, error_msg = tool_error_payload(
                            call_id, fname, f"Error decoding arguments for {fname}: {e}"
                        )
                        yield ChatStreamEvent(content=error_msg)
                        active_history.append(payload)
                        continue

                    try:
                        args = self._filter_args(fname, args)
                        announcement = self._build_tool_announcement_phrase(fname, args)
                        yield ChatStreamEvent(
                            tool_call=ToolCallEvent(
                                tool_name=fname,
                                arguments=args,
                                announcement_phrase=announcement,
                            )
                        )

                        func = self.available_functions[fname]
                        result = (
                            await func(**args)
                            if inspect.iscoroutinefunction(func)
                            else func(**args)
                        )
                    except Exception as e:
                        logger.exception(f"Error executing tool {fname}")
                        result = f"Error: {str(e)}"

                    if isinstance(result, ToolResult):
                        if result.output_schema:
                            yield ChatStreamEvent(
                                reasoning=f"data_{json.dumps(result.output_schema)}"
                            )
                        result = result.content

                    yield ChatStreamEvent(
                        tool_call=ToolCallEvent(tool_name=fname, result=result)
                    )

                    tool_content, vision_message = format_multimodal_content(result)
                    active_history.append(
                        {
                            "tool_call_id": call_id,
                            "role": "tool",
                            "name": fname,
                            "content": tool_content,
                        }
                    )
                    if vision_message:
                        active_history.append(vision_message)

            yield ChatStreamEvent(content="Max tool rounds reached")
        except Exception as e:
            logger.exception(f"Error in chat: {e}")
            yield ChatStreamEvent(content=str(e))

    async def ask_async(
        self,
        user_input: str,
        history: Optional[List[Dict[str, Any]]] = None,
        reasoning: bool = False,
    ) -> str:
        """
        Send a single user message and return the assistant response.

        Args:
            user_input: The user message to send.
            history: Optional conversation history to append to.
            reasoning: Whether to include reasoning content in the stream.

        Returns:
            The assistant response content as a string.
        """
        active_history = history if history is not None else []
        active_history.append({"role": "user", "content": user_input})

        buffer: List[str] = []

        if not self._validate_history(active_history):
            raise ValueError("History is not in the correct format")

        async for event in self.chat(active_history, reasoning=reasoning):
            if event.content:
                buffer.append(event.content)

        response = "".join(buffer)
        active_history.append({"role": "assistant", "content": response})
        return response

    def ask(
        self,
        user_input: str,
        history: Optional[List[Dict[str, Any]]] = None,
        reasoning: bool = False,
    ) -> str:
        """
        Synchronous wrapper for ask_async.

        Raises:
            RuntimeError: If called from within a running event loop.
        """
        try:
            loop = asyncio.get_running_loop()
        except RuntimeError:
            loop = None

        if loop and loop.is_running():
            raise RuntimeError(
                "Agent.ask() cannot be called from a running event loop. "
                "Use `await agent.ask_async(...)` instead."
            )

        return asyncio.run(
            self.ask_async(
                user_input=user_input,
                history=history,
                reasoning=reasoning,
            )
        )

    def show_yourself(self) -> None:
        """Print the agent configuration summary to stdout."""
        info = [
            ("Model", self.model),
            ("Tools", [tool["function"]["name"] for tool in self.tools]),
            (
                "System prompt",
                (self.system_prompt[:100] + "...") if self.system_prompt else None,
            ),
        ]

        for label, value in info:
            cprint(f"{label}: ", "yellow", end="")
            print(value)

    def _build_tool_announcement_phrase(
        self, tool_name: str, arguments: Dict[str, Any]
    ) -> str:
        """Create a human-readable announcement for a tool call."""
        if tool_name in self.tool_objects:
            phrase = str(self.tool_objects[tool_name].announcement_phrase)
            return f"Using {phrase.replace('[args]', str(arguments))}..."

        return f"Using {tool_name}..."

    def _normalize_tool_name(self, name: str) -> str:
        """Strip provider prefixes from tool names."""
        if name.startswith("functions."):
            return name.split(".", 1)[1]
        return name

    def _register_parallel_tool(self) -> None:
        """Register the built-in parallel tool wrapper."""
        tool_name = "multi_tool_use.parallel"
        if tool_name in self.available_functions:
            return

        self.available_functions[tool_name] = self._execute_parallel_tool
        self.tools.append(
            {
                "type": "function",
                "function": {
                    "name": tool_name,
                    "description": "Run multiple tool calls in parallel.",
                    "parameters": {
                        "type": "object",
                        "properties": {
                            "tool_uses": {
                                "type": "array",
                                "items": {
                                    "type": "object",
                                    "properties": {
                                        "recipient_name": {"type": "string"},
                                        "parameters": {"type": "object"},
                                    },
                                    "required": ["recipient_name", "parameters"],
                                },
                            }
                        },
                        "required": ["tool_uses"],
                    },
                },
            }
        )

    def _filter_args(self, tool_name: str, args: Dict[str, Any]) -> Dict[str, Any]:
        """
        Filter arguments to ensure they match the tool's input schema.
        Removes any arguments that are not defined in the schema's properties.
        """
        if tool_name not in self.tool_objects:
            return args

        tool = self.tool_objects[tool_name]
        schema = tool.input_schema

        if not isinstance(schema, dict):
            return args

        # If schema has 'properties', we filter to only allow those keys
        properties = schema.get("properties")
        if properties is not None:
            allowed_keys = set(properties.keys())
            filtered_args = {k: v for k, v in args.items() if k in allowed_keys}

            dropped_keys = set(args.keys()) - set(filtered_args.keys())
            if dropped_keys:
                logger.warning(
                    f"Dropped unexpected arguments for tool {tool_name}: {dropped_keys}"
                )

            return filtered_args

        return args

    async def _execute_parallel_tool(
        self, tool_uses: List[Dict[str, Any]]
    ) -> List[Dict[str, Any]]:
        """Execute multiple tool calls concurrently and return their results."""

        async def run_tool(tool_use: Dict[str, Any]) -> Dict[str, Any]:
            recipient_name = tool_use.get("recipient_name")
            if not recipient_name:
                return {"error": "Missing recipient_name", "tool_use": tool_use}
            params = tool_use.get("parameters") or {}
            tool_name = self._normalize_tool_name(recipient_name)
            params = self._filter_args(tool_name, params)
            if tool_name == "multi_tool_use.parallel":
                return {
                    "recipient_name": recipient_name,
                    "error": "multi_tool_use.parallel cannot invoke itself",
                }
            if tool_name not in self.available_functions:
                return {
                    "recipient_name": recipient_name,
                    "error": f"Function {tool_name} does not exist",
                }
            func = self.available_functions[tool_name]
            try:
                if inspect.iscoroutinefunction(func):
                    result = await func(**params)
                else:
                    result = await asyncio.to_thread(func, **params)
                if isinstance(result, ToolResult):
                    result = result.content
                return {"recipient_name": recipient_name, "result": result}
            except Exception as exc:
                logger.exception(f"Error executing tool {tool_name}")
                return {
                    "recipient_name": recipient_name,
                    "error": f"Error executing {tool_name}: {exc}",
                }

        tasks = [run_tool(tool_use) for tool_use in (tool_uses or [])]
        if not tasks:
            return []
        return await asyncio.gather(*tasks)

    def _validate_history(self, history: List[Dict[str, Any]]) -> bool:
        """Validate the history is in the correct format."""
        for message in history:
            if message.get("role") not in ["system", "user", "assistant", "tool"]:
                raise ValueError(f"Invalid role: {message.get('role')}")
            if message.get("content") is None:
                raise ValueError("Content is required")

        return True



import asyncio
from datetime import datetime
from fury import Agent, create_tool
from apscheduler.schedulers.blocking import BlockingScheduler
from apscheduler.triggers.cron import CronTrigger


def check_time():
    now = datetime.now().isoformat(timespec="seconds")
    return {"current_time": now}


background_tasks = [
    {
        "name": "check_time",
        "prompt": "What time is it right now? Use the check_time tool. If it's in the morning, say 'good morning' and if it's in the afternoon, say 'good afternoon'.",
        "tools": [
            create_tool(
                "check_time",
                "Return the current local time as an ISO-8601 string.",
                check_time,
                "Checking the time...",
                {"type": "object", "properties": {}, "required": []},
                {"type": "object", "properties": {}, "required": []},
            )
        ],
        "schedule": "*/10 * * * * *",
        "reasoning": False,
    }
]


async def run_job(task: dict):
    print(f"\nRunning task: {task['name']}")
    agent = Agent(
        model="unsloth/Qwen3-30B-A3B-GGUF:Q3_K_S",
        system_prompt="You are a helpful assistant. Use the correct tools to answer the user's question.",
        tools=task["tools"],
    )
    buffer = []
    async for event in agent.chat(
        [{"role": "user", "content": task["prompt"]}],
        reasoning=task["reasoning"],
    ):
        if event.content:
            buffer.append(event.content)

    response = "".join(buffer).strip()
    if response:
        print(f"\n[{task['name']}] {response}")


def cron_job(task: dict):
    asyncio.run(run_job(task))


def build_cron_trigger(schedule: str) -> CronTrigger:
    fields = schedule.split()
    if len(fields) == 5:
        return CronTrigger.from_crontab(schedule)
    if len(fields) == 6:
        second, minute, hour, day, month, day_of_week = fields
        return CronTrigger(
            second=second,
            minute=minute,
            hour=hour,
            day=day,
            month=month,
            day_of_week=day_of_week,
        )
    raise ValueError(f"Unsupported cron format: {schedule!r}")


def main():
    scheduler = BlockingScheduler()
    for task in background_tasks:
        scheduler.add_job(
            cron_job,
            trigger=build_cron_trigger(task["schedule"]),
            args=[task],
        )
    scheduler.start()


main()



import json
from typing import Any, Dict, List, Optional

from openai import AsyncOpenAI

DEFAULT_SUMMARY_SYSTEM_PROMPT = (
    "Summarize the conversation for future context using this format:\n"
    "## Goal\n"
    "[What the user is trying to accomplish]\n\n"
    "## Constraints & Preferences\n"
    "- [Requirements mentioned by user]\n\n"
    "## Progress\n"
    "### Done\n"
    "- [x] [Completed tasks]\n\n"
    "### In Progress\n"
    "- [ ] [Current work]\n\n"
    "### Blocked\n"
    "- [Issues, if any]\n\n"
    "## Key Decisions\n"
    "- **[Decision]**: [Rationale]\n\n"
    "## Next Steps\n"
    "1. [What should happen next]\n\n"
    "## Critical Context\n"
    "- [Data needed to continue]\n\n"
    "<read-files>\n"
    "path/to/file1\n"
    "</read-files>\n\n"
    "<modified-files>\n"
    "path/to/file2\n"
    "</modified-files>\n\n"
    "Be concise but include key decisions, filenames, commands, and TODOs."
)


class HistoryManager:
    """Manage conversation history with optional auto-compaction."""

    def __init__(
        self,
        history: Optional[List[Dict[str, Any]]] = None,
        *,
        agent: Optional["Agent"] = None,
        client: Optional[AsyncOpenAI] = None,
        summary_model: Optional[str] = None,
        auto_compact: bool = True,
        context_window: int = 32768,
        reserve_tokens: int = 8192,
        keep_recent_tokens: int = 8000,
        summary_prefix: str = "Summary of previous conversation:",
        summary_system_prompt: str = DEFAULT_SUMMARY_SYSTEM_PROMPT,
    ) -> None:
        self.history: List[Dict[str, Any]] = list(history or [])
        self.context_window = context_window
        self.reserve_tokens = reserve_tokens
        self.keep_recent_tokens = keep_recent_tokens
        self.summary_prefix = summary_prefix
        self.summary_system_prompt = summary_system_prompt
        self.auto_compact = auto_compact

        if agent is not None:
            client = client or agent.client
            summary_model = summary_model or agent.model

        self.client = client
        self.summary_model = summary_model

        if self.auto_compact and (self.client is None or self.summary_model is None):
            raise ValueError(
                "HistoryManager auto compaction requires a client and summary_model "
                "(or an Agent instance)."
            )

    async def add(self, message: Dict[str, Any]) -> List[Dict[str, Any]]:
        """Add a message and auto-compact if configured."""
        self._validate_message(message)
        self.history.append(message)
        if self.auto_compact:
            self.history = await self._compact_history(self.history)
        return self.history

    async def extend(self, messages: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
        """Append multiple messages and auto-compact if configured."""
        for message in messages:
            self._validate_message(message)
        self.history.extend(messages)
        if self.auto_compact:
            self.history = await self._compact_history(self.history)
        return self.history

    def add_nowait(self, message: Dict[str, Any]) -> List[Dict[str, Any]]:
        """Add a message without compacting (sync convenience)."""
        self._validate_message(message)
        self.history.append(message)
        return self.history

    def get_context_usage(self) -> tuple[int, float]:
        tokens = sum(self._estimate_tokens_for_message(msg) for msg in self.history)
        percent = (tokens / self.context_window) * 100 if self.context_window else 0.0
        return tokens, percent

    def _validate_message(self, message: Dict[str, Any]) -> None:
        if message.get("role") not in {"system", "user", "assistant", "tool"}:
            raise ValueError(f"Invalid role: {message.get('role')}")
        if message.get("content") is None:
            raise ValueError("Content is required")

    def _estimate_tokens_for_message(self, message: Dict[str, Any]) -> int:
        role = message.get("role")
        chars = 0

        if role in {"user", "system"}:
            content = message.get("content", "")
            if isinstance(content, list):
                for block in content:
                    if isinstance(block, dict) and block.get("type") == "text":
                        chars += len(block.get("text", ""))
            elif isinstance(content, dict):
                chars += len(json.dumps(content, ensure_ascii=False))
            else:
                chars += len(str(content))
        elif role == "assistant":
            content = message.get("content", "")
            if isinstance(content, list):
                for block in content:
                    if isinstance(block, dict) and block.get("type") == "text":
                        chars += len(block.get("text", ""))
                    elif isinstance(block, dict) and block.get("type") == "toolCall":
                        chars += len(block.get("name", ""))
                        chars += len(
                            json.dumps(block.get("arguments", {}), ensure_ascii=False)
                        )
            else:
                chars += len(str(content))

            tool_calls = message.get("tool_calls")
            if tool_calls:
                chars += len(json.dumps(tool_calls, ensure_ascii=False))
        elif role == "tool":
            content = message.get("content", "")
            if isinstance(content, (dict, list)):
                chars += len(json.dumps(content, ensure_ascii=False))
            else:
                chars += len(str(content))
        else:
            content = message.get("content", "")
            chars += len(str(content))

        return (chars + 3) // 4

    def _should_compact(self, context_tokens: int) -> bool:
        return context_tokens > self.context_window - self.reserve_tokens

    def _find_cut_index(self, messages: List[Dict[str, Any]]) -> int:
        accumulated = 0
        tentative_index = 0

        for i in range(len(messages) - 1, -1, -1):
            accumulated += self._estimate_tokens_for_message(messages[i])
            if accumulated >= self.keep_recent_tokens:
                tentative_index = i
                break
        else:
            return 0

        valid_indices = []
        for idx, msg in enumerate(messages):
            role = msg.get("role")
            if role in {"user", "assistant", "system"}:
                valid_indices.append(idx)

        candidates = [idx for idx in valid_indices if idx <= tentative_index]
        if not candidates:
            return 0
        return max(candidates)

    def _extract_tool_calls(self, message: Dict[str, Any]) -> List[Dict[str, Any]]:
        calls: List[Dict[str, Any]] = []
        tool_calls = message.get("tool_calls")
        if isinstance(tool_calls, list):
            calls.extend([call for call in tool_calls if isinstance(call, dict)])

        content = message.get("content")
        if isinstance(content, list):
            for block in content:
                if isinstance(block, dict) and block.get("type") == "toolCall":
                    calls.append(
                        {
                            "name": block.get("name"),
                            "arguments": block.get("arguments"),
                        }
                    )

        return calls

    def _parse_tool_arguments(self, arguments: Any) -> Dict[str, Any]:
        if isinstance(arguments, dict):
            return arguments
        if isinstance(arguments, str):
            try:
                return json.loads(arguments)
            except json.JSONDecodeError:
                return {}
        return {}

    def _collect_file_ops(
        self, messages: List[Dict[str, Any]]
    ) -> tuple[set[str], set[str]]:
        read_files: set[str] = set()
        modified_files: set[str] = set()

        for msg in messages:
            if msg.get("role") != "assistant":
                continue
            for call in self._extract_tool_calls(msg):
                name = call.get("name")
                args = self._parse_tool_arguments(call.get("arguments"))
                path = args.get("path")
                if not path:
                    continue
                if name == "read":
                    read_files.add(path)
                elif name in {"edit", "write"}:
                    modified_files.add(path)

        return read_files, modified_files

    def _format_message_for_summary(self, message: Dict[str, Any]) -> str:
        role = message.get("role", "unknown")
        if role == "assistant":
            tool_calls = self._extract_tool_calls(message)
            if tool_calls:
                return (
                    f"{role} tool_calls: {json.dumps(tool_calls, ensure_ascii=False)}"
                )

        content = message.get("content", "")
        if isinstance(content, (list, dict)):
            content = json.dumps(content, ensure_ascii=False)

        if role == "tool":
            return f"tool result: {content}"

        return f"{role}: {content}"

    async def _compact_history(
        self, history: List[Dict[str, Any]]
    ) -> List[Dict[str, Any]]:
        if not history:
            return history

        existing_summary = None
        working_history = history

        if (
            history
            and history[0].get("role") == "system"
            and isinstance(history[0].get("content"), str)
            and history[0]["content"].startswith(self.summary_prefix)
        ):
            existing_summary = history[0]["content"][len(self.summary_prefix) :].strip()
            working_history = history[1:]

        context_tokens = sum(
            self._estimate_tokens_for_message(msg) for msg in working_history
        )
        if not self._should_compact(context_tokens):
            return history

        cut_index = self._find_cut_index(working_history)
        if cut_index <= 0:
            return history

        to_summarize = working_history[:cut_index]
        tail = working_history[cut_index:]

        lines = []
        if existing_summary:
            lines.append("Existing summary:")
            lines.append(existing_summary)

        lines.append("Conversation to summarize:")
        for msg in to_summarize:
            lines.append(self._format_message_for_summary(msg))

        read_files, modified_files = self._collect_file_ops(to_summarize)
        if read_files or modified_files:
            lines.append("")
            lines.append("Known file operations (from tool calls in this segment):")
            if read_files:
                lines.append(f"Read files: {', '.join(sorted(read_files))}")
            if modified_files:
                lines.append(f"Modified files: {', '.join(sorted(modified_files))}")

        summary_prompt = "\n".join(lines).strip()
        if not summary_prompt:
            return history

        completion = await self.client.chat.completions.create(
            model=self.summary_model,
            messages=[
                {"role": "system", "content": self.summary_system_prompt},
                {"role": "user", "content": summary_prompt},
            ],
            extra_body={"chat_template_kwargs": {"enable_thinking": False}},
        )
        summary_text = completion.choices[0].message.content or "(summary unavailable)"
        summary_message = f"{self.summary_prefix}\n{summary_text.strip()}"

        return [{"role": "system", "content": summary_message}] + tail


class StaticHistoryManager(HistoryManager):
    """Manage a fixed-size context window without summary compaction."""

    def __init__(
        self,
        *,
        target_context_length: int,
        history: Optional[List[Dict[str, Any]]] = None,
    ) -> None:
        if target_context_length <= 0:
            raise ValueError("target_context_length must be greater than zero")

        super().__init__(
            history=history,
            auto_compact=False,
            context_window=target_context_length,
            reserve_tokens=0,
            keep_recent_tokens=target_context_length,
        )
        self.target_context_length = target_context_length
        self.history = self._fit_to_target(self.history)

    async def add(self, message: Dict[str, Any]) -> List[Dict[str, Any]]:
        self._validate_message(message)
        self.history.append(message)
        self.history = self._fit_to_target(self.history)
        return self.history

    async def extend(self, messages: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
        for message in messages:
            self._validate_message(message)
        self.history.extend(messages)
        self.history = self._fit_to_target(self.history)
        return self.history

    def add_nowait(self, message: Dict[str, Any]) -> List[Dict[str, Any]]:
        self._validate_message(message)
        self.history.append(message)
        self.history = self._fit_to_target(self.history)
        return self.history

    def _fit_to_target(self, messages: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
        fitted: List[Dict[str, Any]] = []
        total_tokens = 0

        for message in reversed(messages):
            message_tokens = self._estimate_tokens_for_message(message)
            if fitted and total_tokens + message_tokens > self.target_context_length:
                break
            if not fitted and message_tokens > self.target_context_length:
                fitted = [message]
                break
            fitted.append(message)
            total_tokens += message_tokens

        return list(reversed(fitted))



from __future__ import annotations

from pathlib import Path
from typing import BinaryIO, Tuple, Union

import numpy as np
import soundfile as sf

AudioSource = Union[str, Path, BinaryIO]


def _to_mono(audio: np.ndarray) -> np.ndarray:
    if audio.ndim == 1:
        return audio
    return np.mean(audio, axis=1)


def _resample_linear(audio: np.ndarray, orig_sr: int, target_sr: int) -> np.ndarray:
    if orig_sr == target_sr:
        return audio
    if orig_sr <= 0 or target_sr <= 0:
        raise ValueError("Sample rates must be positive integers.")

    ratio = target_sr / orig_sr
    target_length = max(int(round(audio.shape[0] * ratio)), 1)
    x_old = np.linspace(0.0, 1.0, num=audio.shape[0], endpoint=False)
    x_new = np.linspace(0.0, 1.0, num=target_length, endpoint=False)
    return np.interp(x_new, x_old, audio).astype(np.float32)


def load_audio(
    source: AudioSource,
    *,
    sr: int | None = None,
    mono: bool = True,
) -> Tuple[np.ndarray, int]:
    audio, sample_rate = sf.read(source, dtype="float32")

    if mono:
        audio = _to_mono(audio)

    if sr is not None and sample_rate != sr:
        audio = _resample_linear(audio, sample_rate, sr)
        sample_rate = sr

    return audio, sample_rate



import asyncio
from datetime import datetime

from fury import Agent, create_tool


def check_time():
    now = datetime.now().isoformat(timespec="seconds")
    return {"current_time": now}


def test_basic_tool_calling():
    tool = create_tool(
        "check_time",
        "Return the current local time as an ISO-8601 string.",
        check_time,
        "Checking the time...",
        {"type": "object", "properties": {}, "required": []},
        {"type": "object", "properties": {}, "required": []},
    )

    agent = Agent(
        model="unsloth/GLM-4.6V-Flash-GGUF:Q8_0",
        system_prompt=(
            "You are a helpful assistant. Use the correct tools to answer the user's question."
        ),
        tools=[tool],
    )

    async def run_chat():
        history = [
            {
                "role": "user",
                "content": "What time is it right now? Use the check_time tool. ",
            }
        ]

        buffer = []
        tool_called = False
        tool_result = None

        async for event in agent.chat(history, reasoning=False):
            if event.tool_call and event.tool_call.tool_name == "check_time":
                tool_called = True
                if event.tool_call.result is not None:
                    tool_result = event.tool_call.result
            if event.content:
                buffer.append(event.content)

        response = "".join(buffer).strip()
        return response, tool_called, tool_result

    response, tool_called, tool_result = asyncio.run(run_chat())
    assert tool_called
    assert tool_result
    assert response



"""
Voice chat example.

Records microphone input, transcribes it with Whisper, streams the reply, and
plays TTS audio back to the user.

Requirements:
  - pip install sounddevice
  - uv add "fury-sdk[tts]" (or pip install "fury-sdk[tts]")
  - espeak (brew install espeak)
"""

import asyncio
import base64
import io

import numpy as np
import sounddevice as sd
import soundfile as sf

from fury import Agent

RECORD_SECONDS = 5.0
INPUT_SAMPLE_RATE = 16000
OUTPUT_SAMPLE_RATE = 24000
REF_AUDIO_PATH = "resources/ref.wav"
REF_TEXT = "Welcome home sir."


def record_audio(duration: float, sample_rate: int) -> str:
    """Record a short WAV clip and return it as base64-encoded bytes."""
    print(f"Recording for {duration:.1f}s...", flush=True)
    frames = int(duration * sample_rate)
    audio = sd.rec(frames, samplerate=sample_rate, channels=1, dtype="float32")
    sd.wait()

    buffer = io.BytesIO()
    sf.write(buffer, audio, sample_rate, format="WAV", subtype="PCM_16")
    return base64.b64encode(buffer.getvalue()).decode("utf-8")


def play_audio(audio: np.ndarray, sample_rate: int) -> None:
    if audio.size == 0:
        return
    sd.play(audio, sample_rate)
    sd.wait()


async def main() -> None:
    agent = Agent(
        model="unsloth/Qwen3-30B-A3B-GGUF:Q3_K_S",
        system_prompt="You are a helpful assistant speaking to a user over voice chat. Be concise, friendly, and helpful.",
    )

    # Warmup tts
    agent.speak(
        text=".",
        ref_text=REF_TEXT,
        ref_audio_path=REF_AUDIO_PATH,
    )

    history = []

    while True:
        input("Press Enter to record: ").strip()

        audio_b64 = record_audio(RECORD_SECONDS, INPUT_SAMPLE_RATE)
        agent.add_voice_message_to_history(history, audio_b64)

        transcript = history[-1]["content"].strip()
        print(f"You said: {transcript}")
        if transcript.lower() in {"q", "quit", "exit"}:
            break

        print("Assistant:")
        reply = ""
        async for event in agent.chat(history):
            if event.content:
                reply += event.content
                print(event.content, end="", flush=True)
        print()

        history.append({"role": "assistant", "content": reply})

        if reply.strip():
            print("Generating TTS audio...")
            audio_chunks = list(
                agent.speak(
                    text=reply,
                    ref_text=REF_TEXT,
                    ref_audio_path=REF_AUDIO_PATH,
                )
            )
            if audio_chunks:
                audio = np.concatenate(audio_chunks)
                play_audio(audio, OUTPUT_SAMPLE_RATE)


if __name__ == "__main__":
    asyncio.run(main())



import asyncio

from fury import Agent, HistoryManager


agent = Agent(
    model="unsloth/GLM-4.6V-Flash-GGUF:Q8_0",
    system_prompt="You are a helpful assistant.",
)

history_manager = HistoryManager(
    agent=agent,
    auto_compact=True,
    context_window=32768,
    reserve_tokens=8192,
    keep_recent_tokens=8000,
)


async def main() -> None:
    while True:
        user_input = input("> ").strip()
        if not user_input:
            continue

        await history_manager.add({"role": "user", "content": user_input})

        buffer = ""
        print()
        async for event in agent.chat(history_manager.history):
            if event.content:
                buffer += event.content
                print(event.content, end="", flush=True)

        await history_manager.add({"role": "assistant", "content": buffer})
        print()


if __name__ == "__main__":
    asyncio.run(main())



"""
Text-to-Speech example.
"""

import wave
import numpy as np

from fury import Agent


def write_wav(path: str, audio: np.ndarray, sample_rate: int = 24000) -> None:
    audio_int16 = (audio * 32767).clip(-32768, 32767).astype(np.int16)
    with wave.open(path, "wb") as wav_file:
        wav_file.setnchannels(1)
        wav_file.setsampwidth(2)
        wav_file.setframerate(sample_rate)
        wav_file.writeframes(audio_int16.tobytes())


def main() -> None:

    agent = Agent(
        model="unsloth/GLM-4.6V-Flash-GGUF:Q8_0",
        system_prompt="You are a helpful assistant.",
    )
    audio_chunks = agent.speak(
        text="Welcome to Fury. The last Agent SDK you'll ever need.",
        ref_text="Welcome home sir.",
        ref_audio_path="resources/ref.wav",
    )

    audio = np.concatenate(list(audio_chunks))
    write_wav("output.wav", audio)


if __name__ == "__main__":
    main()



import wave

import numpy as np

from fury import Agent


def write_wav(path: str, audio: np.ndarray, sample_rate: int = 24000) -> None:
    audio_int16 = (audio * 32767).clip(-32768, 32767).astype(np.int16)
    with wave.open(path, "wb") as wav_file:
        wav_file.setnchannels(1)
        wav_file.setsampwidth(2)
        wav_file.setframerate(sample_rate)
        wav_file.writeframes(audio_int16.tobytes())


def test_tts_generation(tmp_path):
    agent = Agent(
        model="unsloth/GLM-4.6V-Flash-GGUF:Q8_0",
        system_prompt="You are a helpful assistant.",
    )

    audio_chunks = list(
        agent.speak(
            text="Hello from Fury!",
            ref_text="Welcome home sir.",
            ref_audio_path="examples/resources/ref.wav",
        )
    )

    assert audio_chunks

    audio = np.concatenate(audio_chunks)
    output_path = tmp_path / "output.wav"
    write_wav(str(output_path), audio)

    assert output_path.exists()
    assert output_path.stat().st_size > 0



import asyncio

from fury import HistoryManager, StaticHistoryManager


class FakeCompletion:
    def __init__(self, content: str) -> None:
        message = type("Message", (), {"content": content})
        choice = type("Choice", (), {"message": message})
        self.choices = [choice]


class FakeChatCompletions:
    def __init__(self) -> None:
        self.last_request = None

    async def create(self, **kwargs):
        self.last_request = kwargs
        return FakeCompletion("Summary content")


class FakeChat:
    def __init__(self) -> None:
        self.completions = FakeChatCompletions()


class FakeClient:
    def __init__(self) -> None:
        self.chat = FakeChat()


def test_history_manager_auto_compaction():
    client = FakeClient()
    manager = HistoryManager(
        client=client,
        summary_model="fake-model",
        auto_compact=True,
        context_window=60,
        reserve_tokens=10,
        keep_recent_tokens=10,
        summary_prefix="Summary:",
    )

    async def run():
        for idx in range(6):
            await manager.add({"role": "user", "content": f"hello {'x' * 50} {idx}"})
            await manager.add(
                {"role": "assistant", "content": f"reply {'y' * 50} {idx}"}
            )
        return manager.history

    history = asyncio.run(run())

    assert history[0]["role"] == "system"
    assert history[0]["content"].startswith("Summary:")
    assert history[-1]["content"].endswith("5")
    assert client.chat.completions.last_request is not None
    assert len(history) < 12


def test_static_history_manager_fits_initial_history():
    manager = StaticHistoryManager(
        target_context_length=10,
        history=[
            {"role": "user", "content": "a" * 16},  # ~4 tokens
            {"role": "assistant", "content": "b" * 16},  # ~4 tokens
            {"role": "user", "content": "c" * 16},  # ~4 tokens
        ],
    )

    assert len(manager.history) == 2
    assert manager.history[0]["content"] == "b" * 16
    assert manager.history[1]["content"] == "c" * 16


def test_static_history_manager_add_drops_older_messages():
    manager = StaticHistoryManager(
        target_context_length=8,
        history=[
            {"role": "user", "content": "a" * 16},
            {"role": "assistant", "content": "b" * 16},
        ],
    )

    asyncio.run(manager.add({"role": "user", "content": "c" * 16}))

    assert len(manager.history) == 2
    assert manager.history[0]["content"] == "b" * 16
    assert manager.history[1]["content"] == "c" * 16



import asyncio

from fury import Agent


def test_basic_chat_loop():
    agent = Agent(
        model="unsloth/GLM-4.6V-Flash-GGUF:Q8_0",
        system_prompt="You are a helpful assistant.",
    )

    async def run_chat():
        history = [
            {"role": "user", "content": "Hello! Please reply with a short greeting."}
        ]
        buffer = []

        async for event in agent.chat(history, reasoning=False):
            if event.content:
                buffer.append(event.content)

        return "".join(buffer).strip()

    response = asyncio.run(run_chat())
    assert response
